// @ myth - client - macro.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk
//
// The macro class is used by INPUT to translate charecters from the
// keyboard, into commands and arguments.
// The translation include keyboard macros, which is read from the
// file lib/KEY.
//
// To insert a charecter to the macro object, use:
//    put(key)
// The following functions test for speciel keys.
//    bool check(type), argument()
// If there is data ready in the macro object, the following will be true
//    bool ready()
// To extract data from the object, use one of
//    int value(), return numeric value of a key
//    symbol_type symbol(), return an object symbol
//    char key(), return key unchanged
//    direction dir(), return direction of key.
//    confirmation conf(), return confirmation
//    from_client MACRO::command(), return command
//    void edit(const char* string), edit a string

// @@ include

#include "client/macro.h"
#include "common/error.h"
#include "common/parse.h"
#include "common/format.h"

// @@ module KeyTable

// @@@ static const char* KeyName[]

static const char* KeyName[] =
{ "REDRAW", "ESCAPE", "HELP", "LITERAL", "ERASE", "KILL",
  "ACCEPT", "DIRECTION", "CONFIRMATION", "PAGE", NULL };

// @@@ static key_type KeyToType(const char* name)
//
// This look name up in the table KeyName, and return the
// corresponding type.
//
// It should be called like KeyName.type(name), but c++ don't accept
// initialised members of classes yet.

static key_type KeyToType(const char* value)
{
    for(key_type key = 0; key < KEY_END; key++)
	if(strcmp(value, KeyName[key]) == EQUAL)
	    return key;

    return KEY_ERROR;
}

// @@ class key_link

// @@@ bool key_link::pop()

bool key_link::pop()
{
    current++;
    return *current == 0;
}

// @@@ key_link::key_link(char key, key_link* old)

key_link::key_link(char key, key_link* old)
{
    next = old;
    string = new char[2];
    string[0] = key;
    string[1] = 0;
    current = string;
}

// @@@ key_link::key_link(char* macro, key_link* old)

key_link::key_link(char* macro, key_link* old)
{
    next = old;
    string = strcpy(new char[strlen(macro) + 1], macro);
    current = string;
}

// @@@ key_link::~key_link()

key_link::~key_link()
{
    delete string;
}

// @@ class MACRO

// @@@ void MACRO::pop()

void MACRO::pop()
{
    if(stack && (stack->top() == 0 || stack->pop()))
    {
	key_link* next = stack->next;
	delete stack;
	stack = next;
    }
    literal = FALSE;
}

// @@@ void MACRO::put(char key)

void MACRO::put(char key)
{
    if(literal == FALSE && map[key].special[KEY_LITERAL] != ERROR)
	literal = TRUE;
    else
    {
	key_link* head = new key_link(key, stack);
	stack = head;
    }
} 

// @@@ bool MACRO::check(key_type type)

bool MACRO::check(key_type type)
{
    if(!literal && stack && map[stack->top()].special[type] != ERROR)
    {
	pop();
	return TRUE;
    }
    else
	return FALSE;
}

// @@@ int MACRO::value()

int MACRO::value()
{
    int value = stack->top() - '0';
    pop();
    return value;
}

// @@@ symbol_type MACRO::symbol()

symbol_type MACRO::symbol()
{
    if(stack)
    {
	char key = stack->top();
	pop();
	return IsSymbol(key) ? key : SYMBOL_ERROR;
    }
    else
    {
	bug.error("MACRO::symbol()", "Stack underflow");
	return SYMBOL_ERROR;
    }
}

// @@@ char MACRO::key()

char MACRO::key()
{
    if(stack)
    {
	char key = stack->top();
	pop();
	return key;
    }
    else
    {
	bug.error("MACRO::key()", "Stack undeflow");
	return 0;
    }
}

// @@@ direction MACRO::dir()

direction MACRO::dir()
{
    if(stack)
    {
	char key = stack->top();
	pop();
	return map[key].special[KEY_DIRECTION];
    }
    else
    {
	bug.error("MACRO::dir()", "Stack undeflow");
	return DIRECTION_ERROR;
    }
}

// @@@ confirmation MACRO::conf()

confirmation MACRO::conf()
{
    if(stack)
    {
	char key = stack->top();
	pop();
	return map[key].special[KEY_CONFIRMATION];
    }
    else
    {
	bug.error("MACRO::conf()", "Stack undeflow");
	return CONFIRMATION_ERROR;
    }
}

// @@@ from_client MACRO::command()

from_client MACRO::command()
{
    if(stack)
    {
	char key = stack->top();
	pop();
	if(map[key].macro)
	{
	    key_link* head = new key_link(map[key].macro, stack);
	    stack = head;
	}
	return map[key].command;
    }
    else
    {
	bug.error("MACRO::command()", "Stack undeflow");
	return CLIENT_ERROR;
    }
}

// @@@ void MACRO::edit(char* string)

void MACRO::edit(char* string)
{
    int length = strlen(string);

    if(check(KEY_ERASE))
    {
	if(length > 0)
	    string[length - 1] = 0;
    }
    else if(check(KEY_KILL))
	string[0] = 0;
    else
    {
	char c = key();

	if(length < MessageSize - 1 && isprint(c))
	{
	    string[length] = c;
	    string[length + 1] = 0;
	}
    }
}

// @@@ MACRO::MACRO(STRING name)

MACRO::MACRO(STRING name)
{
    for(int i = 0; i < MaxChar; i++)
    {
	map[i].command = CLIENT_ERROR;
	map[i].macro = NULL;
	for(int j = 0; j < KEY_END; j++)
	    map[i].special[j] = ERROR;
    }
    PARSE parse(name);
    
    if(parse.LookAhead(INPUT_COMMENT))
	parse.skip(INPUT_COMMENT);

    while(parse.good())
    {
	if(parse.LookAhead(INPUT_KEY)) // Keyboard Macro
	{
	    parse.skip("\"");
	    char key = parse.GetKey();
	    parse.skip("\"");
	    parse.skip(INPUT_SPACE);
	    map[key].command = CommandToType(parse.GetWord());
	    parse.skip(INPUT_SPACE);
	    if(parse.LookAhead(INPUT_STRING))
	    {
		char* macro = parse.GetString();
		map[key].macro = strcpy(new char[strlen(macro) + 1], macro);
	    }
	}
	else			// Logical Name
	{
	    key_type type = KeyToType(parse.GetWord());
	    parse.skip(INPUT_SPACE);

	    while(parse.good() && parse.LookAhead(INPUT_STRING))
	    {
		char* string = parse.GetString();

		switch(type)
		{
		case KEY_REDRAW:
		case KEY_ESCAPE:
		case KEY_HELP:
		case KEY_LITERAL:
		case KEY_ERASE:
		case KEY_KILL:
		case KEY_ACCEPT:
		case KEY_PAGE:
		    map[*string].special[type] = TRUE;
		    break;
		case KEY_DIRECTION:
		    for(direction dir = 0; dir < DIRECTION_END; dir++)
			map[string[dir]].special[KEY_DIRECTION] = dir;
		    break;
		case KEY_CONFIRMATION:
		    for(confirmation conf = 0; conf < CONFIRMATION_END; conf++)
			map[string[conf]].special[KEY_CONFIRMATION] = conf;
		    break;
		case KEY_END:
		case KEY_ERROR:
		default:
		    bug.error("MACRO::MACRO",
			      format("Unknown logic name %d use %s",
				     type, string));
		}
		parse.skip(INPUT_SPACE);
	    }
	}
	parse.skip(INPUT_COMMENT);
    }
    stack = NULL;
    literal = FALSE;
}

// @@@ MACRO::~MACRO()

MACRO::~MACRO()
{
    for(int i = 0; i < MaxChar; i++)
	delete map[i].macro;
}

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
